######################################################################
#:ClearQuest PerlAPI interface object
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2009. All Rights Reserved.
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#

=head1 NAME

CQCC::CQPerlAPI - Subclass of CQCC::CQuest used to manage
communications with ClearQuest using the CQPerl API interface.

=head1 CLASS DESCRIPTION

This class implements the necessary methods to logon to ClearQuest,
perform queries, and modify records over the CQPerl interface.

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

# Internal Notes: 
# (1) Error checking: Every call to CQPerl must be made within an 
#     eval {} block followed by a call to CQErrorCheck() to handle
#     Errors or Fatals. This will provide a traceback to help identify
#     the source of the problem. 
# (2) Query usage: When we need fields from a number of entities
#     this class uses queries to get the information in batches rather
#     than making repeated calls to GetEntity/GetFieldValue(s), etc. 

######################################################################
#::Package CQCC::CQPerlAPI
######################################################################

package CQCC::CQPerlAPI;
use CQCC::CQuest;
@ISA = qw( CQCC::CQuest );

use strict;
sub Version { return '2.2' };

use CQCC::IOBase;
use CQCC::Parm;
use CQCC::AssocList;
use CQCC::CQSchema;
use CQCC::Timer; 

use CQPerlExt;

require 5.006;
use Socket;
use Getopt::Long;

######################################################################
#::Class Globals 
######################################################################

# Map between operators and coded values
my %boolOp = ("and" => $CQPerlExt::CQ_BOOL_OP_AND,
              "or" => $CQPerlExt::CQ_BOOL_OP_OR
              );

my %compOp = ("eq" => $CQPerlExt::CQ_COMP_OP_EQ,
              "neq" => $CQPerlExt::CQ_COMP_OP_NEQ,
              "lt" => $CQPerlExt::CQ_COMP_OP_LT,
              "lte" => $CQPerlExt::CQ_COMP_OP_LTE,
              "gt" => $CQPerlExt::CQ_COMP_OP_GT,
              "gte" => $CQPerlExt::CQ_COMP_OP_GTE,
              "like" => $CQPerlExt::CQ_COMP_OP_LIKE,
              "not_like" => $CQPerlExt::CQ_COMP_OP_NOT_LIKE,
              "between" => $CQPerlExt::CQ_COMP_OP_BETWEEN,
              "not_between" => $CQPerlExt::CQ_COMP_OP_NOT_BETWEEN,
              "is_null" => $CQPerlExt::CQ_COMP_OP_IS_NULL,
              "is_not_null" => $CQPerlExt::CQ_COMP_OP_IS_NOT_NULL,
              "in" => $CQPerlExt::CQ_COMP_OP_IN,
              "not_in" => $CQPerlExt::CQ_COMP_OP_NOT_IN,
              );

# Schemas and hash tables for managing cc_change_set and _vob_objects
# and multisite mastership entities.

my ($csetSchema, $cvobSchema); 
my (%crChangeSets, %csetVobs, %crDbIDs, %vobDbids);

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=item C<CQCC_MASTER_DB_NAME>

The name of the master database in the data repository for ClearQuest.

=item C<CQCC_DATABASE_SET>

Name of the database set if other than the release name, i.e. 
2001.04.02. 

=back

=cut

######################################################################

sub ClassInit { 
    if (&ParmDefined("CQCC_MASTER_DB_NAME")) { 
        return;
    } 
    &SetParmDefaults(PCLASS=>"CQCC::CQPerlAPI", 
         SRC_CFG=>1, SRC_EV=>1);

    &DefineParm(NAME=>"CQCC_MASTER_DB_NAME", DEFAULT=>"MASTR");

    &DefineParm(NAME=>"CQCC_DATABASE_SET", 
                DEFAULT=>($ENV{"CQ_DBSET"} ? $ENV{"CQ_DBSET"} : ""));

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This class function uses sets up a local environment for testing;
calls IOClassFactory() to get an instance of the desired subclass; 
and invokes the object's SelfTest method for testing.

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;
    &ForceParm("ClearQuestInterface", "CQCC::CQPerlAPI");
    return &CQCC::CQuest::ClassTest(@args);
}

######################################################################
#::function CQErrorCheck
######################################################################

sub CQErrorCheck {
    my ($severity, $caller, $ErrMsg) = @_;

    if ($ErrMsg) {
        my $errMsg = "\n$ErrMsg";
        if (&GetParm("CQCC_DEBUG") > 1) {
            $errMsg .= "\nCall Stack:\n";
            my $i;
            for ($i = 0; $i < 5; $i++) {
                my ($package, $filename, $line, $subr) = caller $i;
                $errMsg .= "($i) $subr, $line\n";
            }
        }
        # Won't return if fatal message, will auto exit
        &IOMessage($severity, [ "CQP_E_ERRCHECK", $caller, $errMsg ]);
        return 1;
    }
    return 0;
} 

######################################################################
#::method Logon

=over 6

=item C<Logon()>

This method attempts to actively logon to the target ClearQuest
database.

=back

=cut


######################################################################

sub Logon {
    my $self = shift;
    my $db = &GetParm("CQCC_DATABASE");
    $self->GetLogonParms();

    my $session = $self->{CQSESSION};

    &IODebugPrint("BASIC",
        "Test Mode: Trying to connect to Perl API\n" .
        "ClearQuest User is: $self->{USER}\n" . 
        "Connecting to DB " . $db . 
        " in " . &GetParm("CQCC_DATABASE_SET") . "\n\n");

    eval {
        $session->UserLogon($self->{USER},$self->{PASS},
             &GetParm("CQCC_DATABASE"),
             &GetParm("CQCC_DATABASE_SET"));
        my $codePageMismatch = 
            $session->GetNameValue("ratl:codePageMismatch");
        if (defined $codePageMismatch && $codePageMismatch ne "") {
            &IOMessage("FATAL", [ "CQP_E_CODE_PAGE_MISMATCH"]);
        }
    };

    if ($@) {
       &IODebugPrint("DETAILED". "Logon Error $@ \n");
       &IOMessage("ERROR", $@);
       return 0;
    }
    else {       
        my $cquser;
        eval {
            $cquser = $session->GetUserLoginName() ;
        };
        &IODebugPrint("DETAILED", "CQUser name is $cquser \n");

        if (defined $cquser) {
        $self->SetCQUser($cquser);
        }

        if (&GetParm("CQCC_DEBUG") > 1) {
            my $entityDefNames = $session->GetEntityDefNames();
            my $name;
            &IODebugPrint("DETAILED", "Entity names in database $db:");
            foreach $name (@$entityDefNames) {
           my $entityDefObj = $self->{CQSESSION}->GetEntityDef($name);
                my $entName = $entityDefObj->GetName();
                &IODebugPrint("DETAILED", "    $entName");
            }
            &IODebugPrint("DETAILED", "\n");
        }
    }
    # Assumes there is only one replica name for a given schema repository

    if (&GetParm("CQCC_MULTISITE") eq "TRUE" 
           && ! defined &GetParm("REPLICA_NAME")) {
        my ($replicaObj, $replicaValue, $replicaName);
        eval {
           $replicaObj = $self->{CQSESSION}->GetLocalReplica();
           $replicaValue = $replicaObj->GetFieldValue("name");
           $replicaName = $replicaValue->GetValue();
        };
        &CQErrorCheck("ERROR", "Logon - Replica Name", $@);
        &ForceParm("REPLICA_NAME", $replicaName);
    } 
    return 1;
}

######################################################################
#::method new

=over 6

=item C<new CQCC::CQPerlAPI()>

This method initializes the object's attributes including the
database and repository names and opens a ClearQuest CQPerl session.

=back

=cut


######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   # Attributes
   my $self;
   # CCperl doesn't support SUPER so use direct reference to parent
   $self = new CQCC::CQuest();

   $self->{CLASSNAME} = "CQCC::CQPerlAPI";
   eval { 
      $self->{CQSESSION} = CQPerlExt::CQSession_Build(); 
   };
   &CQErrorCheck("FATAL", "new", $@);

   # Set up the schema needed for handling secondary entities. 

   if (! defined $csetSchema) { 
       $csetSchema = 
          new CQCC::CQSchema(NAME=>"cc_change_set",
                             TYPE=>"cc_change_set",
                             ENTITY_LIST => "cc_change_set");
       $cvobSchema = 
          new CQCC::CQSchema(NAME=>"cc_vob_object",
                             TYPE=>"cc_vob_object",
                             ENTITY_LIST => "cc_vob_object");

       $cvobSchema->ChangeFieldMap(
            "VOBID" => "vob_family_uuid",
            "OID" => "object_oid",
            "NAME" => "name",
            "MODIFY" => "modify" );
   }

   bless($self, $class);
   return $self;
}

######################################################################
#::method PrintChangeSets

=over 6

=item C<PrintChangeSets(dbName, entityName, @defectIds)>

This method prints out the cc_change_set and cc_vob_objects associated
with one or more defect records. 

=back

=cut

######################################################################

sub PrintChangeSets { 
    my ($self, $dbName, $entity, @defectIDs) = @_;

    my $id;
    my $schema = &GetCQSchemaByEntity($entity);

    $self->_LoadCSetCache($dbName,$schema, $entity, @defectIDs);

    # For each defect, locate it's changeset objects list, then 
    # get the cc_vob_objects that match each entry.

    for $id (@defectIDs) {
        my $csetDbID = $crChangeSets{ $id };
        &IOPrint("\n\nDefect $id cc_change_set list $csetDbID\n");

        if ($csetDbID == 0) {
            next;
        }

        # Process all the cc_vob_objects in the change set
        my $ccVobRef = $csetVobs{ $csetDbID };
        if (defined $ccVobRef) { 
            my @ccVobDbIDlist = @{ $ccVobRef };
            my @ccvDbIDList;

            # Gather list of cc_vob_object dbids for single query
            for (@ccVobDbIDlist) {
               my ($ccvDbID, $name) = split(" ", $_, 2);
               push @ccvDbIDList, $ccvDbID;
            }                
            my @ccVobList = 
                $self->PerformQuery($dbName, "cc_vob_object", 
                    $cvobSchema, 
                    [ "dbid in " . join(",", @ccvDbIDList) ],
                    [ "dbid,<NAME>,<VOBID>,<OID>", "%s=%s!%s!%s" ], 
                    "FALSE");

            for (@ccVobList) {
                IOPrint("\nCC_Vob_object: \n");
                my ($vDbID, $body) = split("=", $_);
                my ($vName, $vfUuid, $oid) = split("!", $body);
                &IOPrint("    Dbid: $vDbID\n" . 
                      "    Pathname: $vName\n" . 
                      "    Vob_Family_UUID: $vfUuid\n" . 
                      "    ObjectID: $oid\n");
            }
        }
    }
}

######################################################################
#::method GetDatabaseList

=over 6

=item C<GetDatabaseList()>

This method returns the list of databases in the repository that
the user is connecting to. 

=back

=cut 

######################################################################

sub GetDatabaseList {
    my $self = shift;
    my ($dbName, $dbdescList);

    eval {
        $dbdescList = 
            $self->{CQSESSION}->GetAccessibleDatabases(
                &GetParm("CQCC_MASTER_DB_NAME"), $self->{USER}, 
                &GetParm("CQCC_DATABASE_SET"));
    };
    &CQErrorCheck("FATAL", "GetDatabaseList", $@);

    # Note - tried to wrap eval/CQErrorCheck around lower section
    # but was getting false errors when GetDatabaseName was wrapped

    if ($dbdescList) {
        my $count = $dbdescList->Count();
        my ($i, @dbNameList);
        
        for ($i = 0; $i < $count; $i++) {
            my $dbdesc = $dbdescList->Item($i);
            push @dbNameList, $dbdesc->GetDatabaseName();
        }
        return @dbNameList;
    }

    return;
}

######################################################################
#::method GetQueryList

=over 6

=item C<GetQueryList(type)>

This method returns the list of queries in the repository that
the user is connecting to. The type argument can be "PUBLIC" 
for only Public queries, "USER" for only user queries or "BOTH"
for both. 

=back

=cut 

######################################################################

sub GetQueryList {
    my ($self,$type) = @_;
    my $result; 
    my (@publicList, @userList);

    eval {
       my $workSpace = $self->{CQSESSION}->GetWorkSpace();

       if ($type eq "PUBLIC" || $type eq "BOTH") { 
           $result = $workSpace->
               GetQueryList($CQPerlExt::CQ_WORKSPACE_PUBLIC_FOLDER);
           if ($result) { 
               @publicList = @{ $result };
           }
       }

       if ($type eq "USER" || $type eq "BOTH") { 
           $result = $workSpace->
               GetQueryList($CQPerlExt::CQ_WORKSPACE_USER_FOLDER);
           if ($result) {
               @userList = @{ $result };
           }
       }
    };
#   Erroneous error reported so skip error check
#    &CQErrorCheck("ERROR", "GetQueryList", $@);

    return (@userList, @publicList);
}

######################################################################
#::method ModifyChangeSets

=over 6

=item C<ModifyChangeSets(entityName, defectKey, operationsHash)>

This internal method modifies an existing cc_change_set list entity, making
additions or deletions based on the entries in the operationsHash list.
Each operation key is the VOList value needed for the assocation and the 
value is either the add or delete operator. 

=back

=cut

######################################################################

sub ModifyChangeSets {
    my ($self, $entityName, $cqDBID, $opHashRef) = @_;
    my $result = 0;
    my $fn = "CQPerlAPI::ModifyChangeSets";

    my ($changeSetEntity, $VOList, $path, $vobObj, @existingVobObjects);
    my $session = $self->{CQSESSION};
    my $schema = &GetCQSchemaByEntity($entityName);
    my $ccs = "cc_change_set";
    my $dbName = substr($cqDBID,0,length($cqDBID)-8);

    # Dereference the operations hash table and return if empty
    my %opHash = %{$opHashRef};
    my @opHashKeys = sort keys %opHash;
    if ($#opHashKeys lt 0) {
        return 1;
    } 
    &IODebugPrint("BASIC", "$fn Processing $entityName $cqDBID");

    # (1) Get or create cc_change_set for editing

    eval {
        my $changeSetID = $crChangeSets{$cqDBID};

        if ($changeSetID eq "") {
           # Create new empty cc_change_set entity
           $changeSetEntity = $session->BuildEntity($ccs);
           $changeSetID = $changeSetEntity->GetDbId();
           $changeSetEntity->Validate();
           $changeSetEntity->Commit();

           # Associate new cc_change_set list with original defect entity
           $self->SetEntityFields( $crDbIDs{$cqDBID}, $schema, $entityName, 
                                   "cc_change_set"=>$changeSetID);
        } else {
           $changeSetEntity = $session->GetEntityByDbId($ccs, $changeSetID);
           my $listItems = 
               $changeSetEntity->GetFieldValue("objects")->GetValueAsList();
           @existingVobObjects = @$listItems;   #@$ dereferences list 
        }
        $changeSetEntity->EditEntity("Add");
    };
    &CQErrorCheck("FATAL", "$fn: Can't create $ccs\n", $@);
    &IODebugPrint("BASIC", "$fn Opened/created cc_change_set record for editing");

    # Add/Remove entries described in operations hash table
    eval {
        for $path (@opHashKeys) {
            my $VOList = $opHash{$path};
            my ($op, $vobid, $oid, $oldOID) = split(",", $VOList);
            my $found = 0;

            # NOTE: Need to handle existing CO associations - lookup PN, etc. 

            if ($op eq "<+>") { 
                # Search existing cached list for pre-existing duplicates
                # skip if found
                for $vobObj (@existingVobObjects) { 
                    if ($vobObj =~/(\S*)\s([\s|\S]*$)/) {
                       my $vobID = $1;
                       my $vobName = $2;
                       if ($vobName eq $path) {
                           $found = 1;
                           last;
                       }
                    }
                }
                # Create new VOB object if needed
                if (! $found) {
                    &IODebugPrint("DETAILED", 
                                  "$fn Creating/adding cc_vob_object $path");
                    my ($assocDbID, $assocKey) = 
                        $self->SetEntityFields(undef, 
                                               $cvobSchema, "cc_vob_object", 
                                               "<VOBID>" => $vobid, 
                                               "<OID>" => $oid, 
                                               "<NAME>" => $path);
                    $changeSetEntity->AddFieldValue("objects", $assocKey);
                 }
            } elsif ($op eq "<->") {
                # Need to search the existing entity list for complete
                # and partial pathname matches
                for $vobObj (@existingVobObjects) { 
                    if ($vobObj =~/(\S*)\s([\s|\S]*$)/) {
                       my $vobID = $1;
                       my $vobName = $2;
                       if ($vobName eq $path) {
                           &IODebugPrint("DETAILED", 
                                         "$fn Deleting cc_vob_object $path");
                           $changeSetEntity->DeleteFieldValue("objects", $vobObj);
                       }
                    }
                }
            }
        }
        &IODebugPrint("DETAILED", "$fn Validating cc_change_set");
        $changeSetEntity->Validate();
        &IODebugPrint("DETAILED", "$fn Committing cc_change_set");
        $changeSetEntity->Commit();
    };
    &CQErrorCheck("FATAL", "$fn($entityName,$cqDBID,...)", $@);
    &IODebugPrint("BASIC", "$fn Finished with $entityName $cqDBID");
    return 1;
}

######################################################################
#::method ProcessBatchOps

=over 6

=item C<ProcessBatchOps(hash)>

This method processes the hash lists built up when the batch information
was processed from a file or other means. 

=back

=cut

######################################################################

sub ProcessBatchOps {
    my ($self,$cqHashRef) = @_;
    my %cqHash = %{$cqHashRef};
    my $result = 1;

    my (%cqKeyHash, $i, $lastDbEntity);
    my @keyList = (sort keys %cqHash);
    for ($i = 0; $i <= $#keyList; $i++) {
        my $cqKey = $keyList[$i];
        my ($entity,$dbID) = split(":",$cqKey);
        my $db = substr($dbID,0,length($dbID)-8);
        my $newDbEntity = join(":", $db, $entity);

        # Set the initial last db-entity value
        if (! defined $lastDbEntity) {
            $lastDbEntity = $newDbEntity;
        }
        # When using same DB and entity, build up a list for processing
        if ($newDbEntity eq $lastDbEntity) {
            $cqKeyHash{$cqKey} = $dbID;
        } 
        # If DB/Entity changed, process the list, reset the accumulators
        if ($newDbEntity ne $lastDbEntity) {
            my $r = $self->_PBOinternal($lastDbEntity, $cqHashRef, \%cqKeyHash);
            if ($r == 0) {
                $result = 0;
            }
            # Reset for next batch

            %cqKeyHash = ();
            $cqKeyHash{$cqKey} = $dbID;
            $lastDbEntity = $newDbEntity;
        }
    }
    # Process the final batch 
    my $r = $self->_PBOinternal($lastDbEntity, $cqHashRef,\%cqKeyHash);
    if ($r == 0) {
        $result = 0;
    }   
    return $result;
}

# Internal method used as part of ProcessBatchOps only

sub _PBOinternal {
    my ($self,$dbEntity,$cqHashRef,$cqKeyHashRef) = @_;
    my %cqKeyHash = %{$cqKeyHashRef};
    my @cqKeyKeys = sort keys %cqKeyHash;
    my @cqKeyIDs = sort values %cqKeyHash;
    my $result = 0;
    my $init_err_count = &IOstd->{ERR_COUNT};

    if ($#cqKeyKeys eq -1) {
        return 1;
    } 

    my ($db,$entity) = split(":", $dbEntity);
    
    if ($self->SetDBEntity($db,$entity)) {
        my %cqHash = %{$cqHashRef};
        # Load a cache of existing CQ change_set information 
        my $schema = &GetCQSchemaByEntity($entity);
        # Clear existing caches
        %crChangeSets = ();
        %crDbIDs = ();

        # Query 1 for defects' objects fields (cc_change_set dbids)
        my @data = $self->PerformQuery($db, $entity, $schema,
            [ "<ID> in " . join(",", @cqKeyIDs) ],
            [ "<ID>,dbid,cc_change_set", "%s %s %s" ], "FALSE");

        if (&IOstd->{ERR_COUNT} > $init_err_count) {
            return 0;
        }
        for (@data) {
             my ($defectID, $defectDbID, $objectsID) = split(" ", $_);
             if ($objectsID > 0) {
                 $crChangeSets{$defectID} = $objectsID;
             }
             $crDbIDs{$defectID} = $defectDbID;
        }

        $result = 1;
        # Make changes to the specific defects on the list 
        my $cqKey;
        for $cqKey (@cqKeyKeys) {
            my $dbID = $cqKeyHash{$cqKey};
            if ($self->ModifyChangeSets($entity,$dbID,$cqHash{$cqKey}) == 0) {
                $result = 0;
            }
        }
    }
    return $result;
}

######################################################################
#::method SetEntityFields

=over 6

=item C<SetEntityFields(dbid, schema, entityName, fieldValues)>

This method creates a new entity of type "entityName" (if dbID is
undefined) or modifies an existing one if dbID is provided. It sets
the entity's field values to the paired list provided (mapping the
names if needed), validates and commits the entry and returns the dbid
and key of the new entity.

(dbid,k) = obj->SetEntityFields(dbid,schema,entityName,fieldValues);

For handling fields that take list values use SetEntityListField
method below. The two methods can be used on the same entity if needed
though this will make two separate editing passes. 

=back

=cut

######################################################################

sub SetEntityFields {
    my ($self, $dbID, $schema, $entityName, @fieldList) = @_;

    my ($new_key, $fieldID, $value, $new_entity);
    
    ($dbID,$new_entity) = 
        $self->_GetEntity($dbID, $schema, $entityName, "<MODIFY>");

    IODebugPrint("BASIC",
        "CQPerlAPI::SetEntityFields db($dbID) entity($entityName)\n");

    eval {
        while (@fieldList) { 
            $fieldID = shift @fieldList;
            $value = shift @fieldList;
            if (substr($fieldID,0,1) eq "<") {
                ($fieldID) = $schema->EvalItem($fieldID);
            }
            &IODebugPrint("BASIC", "    Field $fieldID = $value");
            $new_entity->SetFieldValue($fieldID, $value);
        }
        $new_key = $new_entity->GetDisplayName();
        $new_entity->Validate();
        $new_entity->Commit();
    };
    &CQErrorCheck("FATAL", "SetEntityFields", $@);

    return ($dbID, $new_key);
}

######################################################################
#::method SetEntityListField

=over 6

=item C<SetEntityListField(dbid, schema, entityName, fieldName, 
operation, \@valueList)>

This method creates a new entity of type "entityName" (if dbID is
undefined) or modifies an existing one if dbID is provided. It only
modifies a single field with the name "fieldName" based on the
operation ("Add", "Remove", "Clear") and the valueList provided.  If
the operation is "Add", the value list is appended to the existing
value list; if "Remove", the value list is deleted from the existing
list; and if "Clear", all existing values are deleted. If necessary
the "fieldName" is mapped using the schema.  The method then validates
the changes and commits the entry and returns the dbid and key of the
new entity.

(dbid,k) = $obj->SetEntityListField(dbid, schema, entityName, 
     fieldName, operation, \@valueList)

Use the SetEntity method to modify non-list fields. 

=back

=cut

######################################################################

sub SetEntityListField {
    my ($self, $dbID, $schema, $entityName, $fieldID, $operation,
       @valueList) = @_;

    my ($new_key, $fieldName, $new_entity);
    
    ($dbID, $new_entity) = 
        $self->_GetEntity($dbID, $schema, $entityName, $operation);

    $fieldName = $fieldID;
    if (substr($fieldName,0,1) eq "<") {
        ($fieldName) = $schema->EvalItem($fieldName);
    }

    eval {
        if ($operation eq "Clear") {
            my $fieldInfo = $new_entity->GetFieldValue($fieldName);
            my @valueList = $fieldInfo->GetValueAsList();
            $operation = "Remove";
        }
    
        if ($operation eq "Add") {
            for (@valueList) {
                $new_entity->AddFieldValue($fieldName, $_);
            }
        } elsif ($operation eq "Remove") {
            for (@valueList) {
                $new_entity->DeleteFieldValue($fieldName, $_);
            }
        }
        
        $new_key = $new_entity->GetDisplayName();
        $new_entity->Validate();
        $new_entity->Commit();
    };
    &CQErrorCheck("FATAL",
        "SetEntityListField($fieldID)", $@);

    return ($dbID, $new_key);
}

######################################################################
#::method _EditAssocs

=over 6

=item C<_EditAssocs(database,entity,$VOList,add,del)>

This internal method implements the EditAssocs functionality for
a single database and entity. The add and del association lists are
expected to be homogeneous since we can only interact with one
database and one entity type at a time. The EditAssocs method is
responsible for splitting up associations into homogeneous lists. 

It must be called thru the CQuest method, EditAssocs which will 
break up the association lists into unique database/entity 
combinations so that this method only handles one combination at a 
time. 

This method calls _LoadCSetCache to speed up getting dbids for 
cc_change_set, cc_vob_object entities as well as the OIDS when
required for the files.  This may be overkill in
some situations but should scale up well for larger numbers of 
associations per changeset.

=back

=cut

######################################################################

sub _EditAssocs {
    my ($self,$dbName,$entity,$VOList,$addAssocs,$delAssocs) = @_;
    my ($cqname, $vobid, $oid, $oldOID) = split(",", $VOList);
    my $result = 1;
    my $parmSet = 0;
    my (@addIDs, @delIDs, $id, $hashOID, $dupOID, $vobs, $delimiter);
    my $schema = &GetCQSchemaByEntity($entity);
    my $init_err_count = &IOstd->{ERR_COUNT};
    my %deletedKeys;  # Hash of elements deleted already

    # For efficiency and scalability, we do cache up the results of
    # two queries as hash tables to get the dbids we need for 
    # cc_change_sets and cc_vob_objects.     Sometimes a third
    # query is needed, this query has the OIDS and under certain
    # conditions we need to match the OIDS.

    if ($addAssocs) {
        @addIDs = $addAssocs->GetBugList();
        if (@addIDs && ($parmSet == 0)) {
            if (&GetParm("CQCC_QUERY_OID") eq "TRUE") {
                $parmSet = 1;
            }
        }
    } 
    if ($delAssocs) {
        @delIDs = $delAssocs->GetBugList();
        if (@delIDs && ($parmSet == 0)) {
            &ForceParm("CQCC_QUERY_OID", "TRUE");
            $parmSet = 1;
        }
    } 

    if (&GetParm("CQCC_DEBUG") > 1) { 
        &IODebugPrint("DETAILED", "CQPerlAPI::_EditAssocs:\n" . 
                      "   Add: " . join(",",@addIDs) . "\n" . 
                      "   Del: " . join(",",@delIDs) . "\n" . 
                      "   _LoadCSetCache to get cc_change_sets\n");
    }
    $self->_LoadCSetCache($dbName,$schema, $entity, @addIDs, @delIDs);
    if (&IOstd->{ERR_COUNT} > $init_err_count) {
        return 0;
    }

    $delimiter = $ENV{"CLEARCASE_XN_SFX"};
    if (! defined $delimiter) {
        $delimiter = "@@";
    }

    # Delete existing association if found
    # If no change_set record then ignore
    # Loop over delIDs for each defect
    #   delete assoc from objects list 
    #   update hash table

    if (@delIDs) {
        # Because we may be deleting checkout associations during
        # checkin, we might be working with $cqname being an XPN
        # and having to delete the checkout names which used PN.
        # So if the the existing CQ assocation is a PN name, 
        # then check it against the PN part of $cqname only or we 
        # won't match. 

        # Note - @@ may not always be delimiter but can't always rely
        # on ClearCase env

        my ($cqname_PN) = split($delimiter, $cqname);

        for $id (@delIDs) {
            my $csetDbID = $crChangeSets{ $id };
            if ($csetDbID ne "0") {
                $vobs = $csetVobs{ $csetDbID };
                if ($vobs) {
                    my @vobsList = @{ $vobs };
                    my $vob;
                    for $vob (@vobsList) {
                        my ($dbid,$p) = split(" ", $vob, 2);
                        my ($p_pn, $p_branch) = split($delimiter, $p);
                        $hashOID = $vobDbids  { $dbid};
                                                
                        # Note: Don't delete VOB record because of 
                        # possible use by other entries

                        if ((defined $p_branch && $p eq $cqname) ||
                            (! defined $p_branch && 
                               $p eq $cqname_PN) ||
                            (($parmSet) &&
                            ($hashOID eq $oldOID))) {
                            $self->SetEntityListField( $csetDbID,
                                $csetSchema, "cc_change_set",
                                "objects", "Remove", $vob);
                                
                            # Remember the deleted element
                            $deletedKeys{"$id:$vob"} = 1;
                        }
                    }
                }
            }
        }
    }

    if (@addIDs) {  
        #Make association if possible
        #    Create new association entity with name, vobid, 
        #    Loop over addlist for each defect
        #        add assoc to objects list and add to defect if new
        #        update hash tables (?)

        my ($assocKey, $assocDbID) = 
            $self->SetEntityFields(undef, $cvobSchema,
                 "cc_vob_object", "<VOBID>" => $vobid, 
                 "<OID>" => $oid, "<NAME>" => $cqname);

        for $id (@addIDs) {
            $dupOID = 0;
            my $csetDbID = $crChangeSets{ $id };
            # check to see if OID is already there
            # if it is do not add it #dupOID = 1
            if (($parmSet) && ($csetDbID ne "0")) { 
                $vobs = $csetVobs{ $csetDbID };
                if ($vobs) {
                    my @vobsList = @{ $vobs };
                    my $vob;
                    for $vob (@vobsList) {
                        my ($dbid,$p) = split(" ", $vob, 2);
                        $hashOID = $vobDbids  { $dbid};
                        if (($hashOID eq $oid) && (! $deletedKeys{"$id:$vob"})) {
                            $dupOID = 1;
                            last;
                        }
                    }
                }
            }

            if (!$dupOID) {
                my ($newDbID,$newKey) = 
                    $self->SetEntityListField($csetDbID,
                        $csetSchema, "cc_change_set",
                        "objects", "Add", $assocDbID);
                if (! defined $csetDbID) {
                    $crChangeSets{ $id } = $newDbID;
                    $self->SetEntityFields( $crDbIDs{ $id },
                        $schema, $entity, "cc_change_set"=>$newDbID);
                }
           }
        }
    }

    if (&GetParm("CQCC_DEBUG") > 1) {
       &IODebugPrint("DETAILED", "CQPerl::_EditAssocs - " . 
                     "reload change set cache to see results\n");
       $self->_LoadCSetCache($dbName,$schema, $entity, @addIDs, 
                             @delIDs);
    }

    if (&IOstd->{ERR_COUNT} > $init_err_count) {
        return 0;
    }
    return $result;
}

######################################################################
#::method _PerformQuery

=over 6

=item C<_PerformQuery(db,entity,schema,query,results,resultsFormat,
returnType, queryType)>

This internal method is called by PerformQuery to actually query the
ClearQuest database for records matching the query description.

returnType controls the format of the results - STRING returns
formatted strings and ARRAY returns each record as a list of fields. 

queryType identifies what kind of query is being used - LOCAL means
the query was defined in the config.pl file and CQ means it is a named
query defined in the ClearQuest database. 

=back

=cut

######################################################################

sub _PerformQuery {
    my ($self, $dbname, $entity, $schema, $query, $resultsSet, 
       $resultsFormat, $returnType, $queryType) = @_;
    my ($field, $status, @bugdesc, $filter, $operator,
        $querydef, @display_fields, $CQresultSet);
    my $n = 0;
    my $firstColumn = 0;

    # Working with config file defined query so convert filters
    # to a query definition. 

    if ($queryType eq "LOCAL") {
        my @filters = @{ $query };  # Must be an array reference
        # NOTE: Problems with eval/CQCheckError showing false errors
        # Breaking them up seemed to help. 

        eval {
            $querydef = $self->{CQSESSION}->BuildQuery($entity);
            @display_fields = split(",", $resultsSet);
        
            for $field (@display_fields) {
                $querydef->BuildField($field);
            }
            $operator = 
               $querydef->BuildFilterOperator($boolOp{"and"});
        };
        &CQErrorCheck("FATAL", "PerformQuery(1)", $@);

        eval {
            for (@filters) {
               my ($subject,$verb,$object) = split(" ",$_,3);
               my @args = split(",",$object);
               my $op = $compOp{ lc($verb)};
    
               if (!defined $op) {
                   &IOMessage("FATAL", 
                      ["CQP_E_QUERY_INVALID", $verb,
                       join("\n  ",@filters)]);
               }
               $operator->BuildFilter($subject, $op, \@args);
            }
        };
        &CQErrorCheck("FATAL", "PerformQuery(2)", $@);
    }    
    # Else working with named query in the CQ database ("CQ")
    # so we need to get the query definition from ClearQuest

    else {
        # Note: CQErrorCheck not working
        my $workSpace = $self->{CQSESSION}->GetWorkSpace();
        $querydef = $workSpace->GetQueryDef($query);
        $firstColumn = 1; 
        if (! defined $querydef) { 
            return ();
        } 

        # Need to make sure the first field is ID in user defined
        # queries. The ID field name itself can't be changed 
        # because it is provided by the CQ packages. 

        my $fieldDefs = $querydef->GetQueryFieldDefs();
        if ($fieldDefs) {
            my $field1 = $fieldDefs->Item(1);
            if ($field1) {
                my $fieldName1 = $field1->GetFieldPathName();
                if (uc($fieldName1) ne "ID") {
                    &IOMessage("ERROR", 
                        ["TCQCC_E_ID_MUST_BE_FIRST",$query]);
                    return ();
                }
            }
        }
    }

    # For both query types, run the query and retrieve results

    eval {
        $CQresultSet = $self->{CQSESSION}->BuildResultSet($querydef);
        $CQresultSet->Execute;
        $status = $CQresultSet->MoveNext;

        if (! $status eq $CQPerlExt::CQ_SUCCESS){ 
            return; 
        } 
 
        my $columnCount = $CQresultSet->GetNumberOfColumns();
        my $resultsEntity = 
            $CQresultSet->LookupPrimaryEntityDefName();
        if (! (uc($resultsEntity) eq uc($entity))) {
            &IOMessage("WARN", ["CQP_E_WRONG_ENTITY", $entity,
                       $resultsEntity]);
            return ();
        } 

        # For the ARRAY type, return the column headings as the 
        # first record entry. 

        if ($returnType eq "ARRAY") { 
            my (@tempFld,$i);
            for ($i = $firstColumn; $i < $columnCount; $i++) {
                $tempFld[$i-$firstColumn] = 
                    $CQresultSet->GetColumnLabel($i+1);
            }
            $bugdesc[$n++] = \@tempFld;
        }
        while ($status eq $CQPerlExt::CQ_SUCCESS){
            my (@fld,$i);
            for ($i = $firstColumn; $i < $columnCount; $i++) {
                $fld[$i-$firstColumn] = 
                    $CQresultSet->GetColumnValue($i+1);
            }
            if ($returnType eq "STRING") { 
                $bugdesc[$n++] = sprintf($resultsFormat, @fld);
            }
            else {
                # Only try to format Locally defined queries
                # Split the results and loop thru the array and apply the format to it's respective field
                my @tempFld;
                if ($queryType eq "LOCAL") {
                        my $column = 0;
                        my @resultsFormatArray = split(/\s/, $resultsFormat);
                        foreach my $format (@resultsFormatArray) {
                                $tempFld[$column] = sprintf($format, $fld[$column]);
                                $column++;
                        }
                }
		else {
			@tempFld = @fld;
		}
                $bugdesc[$n++] = \@tempFld;
            }
            $status = $CQresultSet->MoveNext;
        }
    };
    &CQErrorCheck("FATAL", "PerformQuery(3)", $@);

    if ($returnType eq "STRING") {
        return sort @bugdesc;
    }
    return @bugdesc;
}

######################################################################
#::method _LoadCSetCache

=over 6

=item C<_LoadCSetCache(DB, Schema, entity, @ids)>

This method caches up change set information for the defect records of
immediate interest. It runs two queries whose output is used to (1)
build up a hash map from defects to their object (cc_change_set
pointer) fields and (2) a hash map from change_sets to their lists of
cc_vob_object keys. These hash tables are then used by EditAssocs
to more efficiently make changes. 

=back

=cut

######################################################################

sub _LoadCSetCache {
    my ($self, $dbName, $schema, $entityName, @idList) = @_;

    my (@data, @ccVobRec);
    my $parmSet = 0;

    # Clear existing caches

    %crChangeSets = ();
    %crDbIDs = ();
    %csetVobs = ();
    %vobDbids = ();

    # Check to see if we are doing OID query

    if (&GetParm("CQCC_QUERY_OID") eq "TRUE") {
        $parmSet = 1;
    }

    # Query 1 for defects' objects fields (cc_change_set dbids)    

    @data = $self->PerformQuery($dbName, $entityName, $schema, 
         [ "<ID> in " . join(",", @idList) ],
         [ "<ID>,dbid,cc_change_set", "%s %s %s" ], "FALSE");

    for (@data) {
        my ($defectID, $defectDbID, $objectsID) = split(" ", $_);
        if ($objectsID > 0) {
            $crChangeSets{$defectID} = $objectsID;
        }
        $crDbIDs{$defectID} = $defectDbID;
    }

    # Query 2 for changesets contents accessible by changeset dbid
    # Note: LoadCSetCache results set defined in ClassInit

    my $csets = join(",", values %crChangeSets);
    if ($csets) {
        @data = $self->PerformQuery($dbName, "cc_change_set",
            $csetSchema,
            [ "dbid in $csets" ], [ "dbid,objects", "%s %s" ], 
            "FALSE");

        my ($cset_key,@vobList);

        for (sort @data) {
            my ($key, $vob, $path) = split(" ", $_, 3);
            if (! (defined $path && $path ne "")) {  # Skip empty results
		next;
            }
            push @ccVobRec, $vob;
            my $value = $vob . " " . $path;
            if (! defined $cset_key) {
               $cset_key = $key;
            } elsif ($key ne $cset_key) {
               # Store references to temp arrays so they 
               # don't all reference the same location
               my @localVobList = @vobList;
               $csetVobs{$cset_key} = \@localVobList;
               $cset_key = $key;
               @vobList = ();
            }
            push @vobList, $value;
        }
        if (defined $cset_key) { 
            $csetVobs{$cset_key} = \@vobList;
        }
    }
   
    # Query 3 for OIDS accessible by vob dbid
    # Note: LoadCSetCache results set defined in ClassInit
   
    if ($parmSet == 1) {
       my $ccVobDbidList = join(",", @ccVobRec);
       if ($ccVobDbidList) {
             @data = $self->PerformQuery($dbName, "cc_vob_object",
             $cvobSchema,
             [ "dbid in $ccVobDbidList" ],
             [ "dbid,<NAME>,<VOBID>,<OID>", "%s,%s,%s,%s" ],
             "FALSE");

            for (@data) {
                my  ($vobDbID, $vobName, $vobid, $vobOid) = split(",", $_ , 4);
                if ($vobOid) {
                    $vobDbids{$vobDbID} = $vobOid;
                }
            }
        }
     }
 
    return;
}

# Internal routine used in testing
sub _RetrieveCSetCache {
    my ($self) = @_;
    return (\%crChangeSets, \%crDbIDs, \%csetVobs);
}

######################################################################
#::method _GetEntity

=over 6

=item C<_GetEntity(dbid, schema, entityName)>

This internal method creates a new entity of type "entityName" (if
dbID is undefined) or modifies an existing one if dbID is provided.
It is only used by other methods that make field changes then 
validate and commit the information. 

(dbid,entity) = obj->_GetEntity(dbid, schema, entityName);

=back

=cut

######################################################################

sub _GetEntity {
    my ($self, $dbID, $schema, $entityName, $operation) = @_;
    my ($new_entity);

    # Create entity if no $dbID is provided

    if (! defined $dbID) {
        eval { 
             $new_entity = 
                 $self->{CQSESSION}->BuildEntity($entityName);
             $dbID = $new_entity->GetDbId();
        };
        &CQErrorCheck("FATAL","_GetEntity (create)", $@);

    } else {  # Else attempt to get existing entiyt
        eval {
            $new_entity = 
                $self->{CQSESSION}->GetEntityByDbId($entityName, 
                                                    $dbID);
            my $modify = $operation;
            if (substr($operation,0,1) eq "<") {
                ($modify) = $schema->EvalItem($operation);
            }
            $self->{CQSESSION}->EditEntity($new_entity, $modify);
        };
        &CQErrorCheck("FATAL", "_GetEntity (modify)", $@);
    }
    return ($dbID, $new_entity);
}

######################################################################
#::Tail
######################################################################

1;
